@@ -1,4 +1,4 @@
-__version__ = '0.4.0'
+__version__ = '0.5.0'
 
 # Set up logging
 import logging
@@ -10,5 +10,11 @@ logging.getLogger('newspaper').disabled = True
 logging.getLogger('newspaper').propagate = False
 logging.getLogger('jieba').disabled = True
 logging.getLogger('jieba').propagate = False
+logging.getLogger('numexpr').disabled = True
+logging.getLogger('numexpr').propagate = False
+logging.getLogger('bs4').disabled = True
+logging.getLogger('bs4').propagate = False
+logging.getLogger('transformers').disabled = True
+logging.getLogger('transformers').propagate = False
 
 from .dataset import MuminDataset  # noqa
