@@ -3,7 +3,8 @@
 from pathlib import Path
 from typing import Union, Dict, Tuple, List, Optional
 import pandas as pd
-from pandas.errors import PerformanceWarning
+import zipfile
+import io
 import numpy as np
 import logging
 import requests
@@ -15,6 +16,7 @@ from tqdm.auto import tqdm
 import warnings
 import json
 from functools import partial
+from shutil import rmtree
 
 from .twitter import Twitter
 from .article import process_article_url
@@ -42,7 +44,7 @@ class MuminDataset:
             The Twitter bearer token.
         size (str, optional):
             The size of the dataset. Can be either 'small', 'medium' or
-            'large'. Defaults to 'large'.
+            'large'. Defaults to 'small'.
         include_replies (bool, optional):
             Whether to include replies and quote tweets in the dataset.
             Defaults to True.
@@ -65,8 +67,8 @@ class MuminDataset:
             The HuggingFace Hub model ID to use when embedding images. Defaults
             to 'google/vit-base-patch16-224-in21k'.
         dataset_path (str, pathlib Path or None, optional):
-            The path to the folder where the dataset should be stored. If None
-            then the dataset will be stored at './mumin-<size>.hdf'. Defaults
+            The path to the file where the dataset should be stored. If None
+            then the dataset will be stored at './mumin-<size>.zip'. Defaults
             to None.
         verbose (bool, optional):
             Whether extra information should be outputted. Defaults to True.
@@ -83,6 +85,9 @@ class MuminDataset:
         image_embedding_model_id (str): The model ID used for embedding images.
         nodes (dict): The nodes of the dataset.
         rels (dict): The relations of the dataset.
+        compiled (bool): Whether the dataset has been compiled.
+        verbose (bool): Whether extra information should be outputted.
+        download_url (str): The URL to download the dataset from.
 
     References:
         - [1] Nielsen and McConville: _MuMiN: A Large-Scale Multilingual
@@ -90,8 +95,8 @@ class MuminDataset:
               Network Posts_ (2021)
     '''
     download_url: str = ('https://github.com/CLARITI-REPHRAIN/mumin-build/raw/'
-                         'c5276bab7e9421daaa4e6acb7d7ee48cd966b329/'
-                         'data/mumin.hdf')
+                         'c6700f77c4c6cffcdb0bbfad21feb40c153f9a75'
+                         '/data/mumin.zip')
     _node_dump: List[str] = ['claim', 'tweet', 'user', 'image', 'article',
                              'hashtag', 'reply']
     _rel_dump: List[Tuple[str, str, str]] = [
@@ -114,7 +119,7 @@ class MuminDataset:
 
     def __init__(self,
                  twitter_bearer_token: str,
-                 size: str = 'large',
+                 size: str = 'small',
                  include_replies: bool = True,
                  include_articles: bool = True,
                  include_images: bool = True,
@@ -140,7 +145,7 @@ class MuminDataset:
         self.rels: Dict[Tuple[str, str, str], pd.DataFrame] = dict()
 
         if dataset_path is None:
-            dataset_path = f'./mumin-{size}.hdf'
+            dataset_path = f'./mumin-{size}.zip'
         self.dataset_path = Path(dataset_path)
 
         # Set up logging verbosity
@@ -179,8 +184,11 @@ class MuminDataset:
         self._download(overwrite=overwrite)
         self._load_dataset()
 
+        # Variable to check if dataset has been compiled
+        compiled = self.compiled or ('text' in self.nodes['tweet'].columns)
+
         # Only compile the dataset if it has not already been compiled
-        if 'text' not in self.nodes['tweet'].columns:
+        if not compiled:
 
             # Shrink dataset to the correct size
             self._shrink_dataset()
@@ -202,12 +210,16 @@ class MuminDataset:
             self._filter_node_features()
             self._filter_relations()
 
+            # Set datatypes
+            self._set_datatypes()
+
         # Remove unnecessary bits
         self._remove_auxilliaries()
         self._remove_islands()
 
         # Save dataset
-        self._dump_dataset()
+        if not compiled:
+            self._dump_dataset()
 
         # Mark dataset as compiled
         self.compiled = True
@@ -250,6 +262,33 @@ class MuminDataset:
                             pbar.update(len(data))
                             f.write(data)
 
+            logger.info('Converting dataset to less compressed format')
+
+            # Open the zip file containing the dataset
+            data_dict = dict()
+            with zipfile.ZipFile(self.dataset_path,
+                                 mode='r',
+                                 compression=zipfile.ZIP_DEFLATED) as zip_file:
+
+                # Loop over all the files in the zipped file
+                for name in zip_file.namelist():
+
+                    # Extract the dataframe in the file
+                    byte_data = zip_file.read(name=name)
+                    df = pd.read_pickle(io.BytesIO(byte_data),
+                                        compression='xz')
+                    data_dict[name] = df
+
+            # Overwrite the zip file in a less compressed way, to make io
+            # operations faster
+            with zipfile.ZipFile(self.dataset_path,
+                                 mode='w',
+                                 compression=zipfile.ZIP_STORED) as zip_file:
+                for name, df in data_dict.items():
+                    buffer = io.BytesIO()
+                    df.to_pickle(buffer, protocol=4)
+                    zip_file.writestr(name, data=buffer.getvalue())
+
         return self
 
     def _load_dataset(self):
@@ -269,37 +308,48 @@ class MuminDataset:
         self.nodes = dict()
         self.rels = dict()
 
-        with pd.HDFStore(self.dataset_path) as hdf:
-            ntypes = [name.replace('/', '')
-                      for name in hdf.keys() if '_' not in name]
-            etypes = [name.replace('/', '')
-                      for name in hdf.keys() if '_' in name]
+        # Open the zip file containing the dataset
+        with zipfile.ZipFile(self.dataset_path,
+                             mode='r',
+                             compression=zipfile.ZIP_STORED) as zip_file:
 
-            for ntype in ntypes:
-                self.nodes[ntype] = hdf[ntype].copy()
+            # Loop over all the files in the zipped file
+            for name in zip_file.namelist():
 
-            for etype in etypes:
-                splits = etype.split('_')
-                src = splits[0]
-                tgt = splits[-1]
-                rel = '_'.join(splits[1:-1])
-                self.rels[(src, rel, tgt)] = hdf[etype].copy()
+                # Extract the dataframe in the file
+                byte_data = zip_file.read(name=name)
+                df = pd.read_pickle(io.BytesIO(byte_data))
 
-        # Ensure that claims are present in the dataset
-        if 'claim' not in self.nodes.keys():
-            raise RuntimeError('No claims are present in the file!')
+                # If there are no underscores in the filename then we assume
+                # that it contains node data
+                if '_' not in name:
+                    self.nodes[name.replace('.pickle', '')] = df.copy()
 
-        # Ensure that tweets are present in the dataset, and also that the
-        # tweet IDs are unique
-        if 'tweet' not in self.nodes.keys():
-            raise RuntimeError('No tweets are present in the file!')
-        else:
-            tweet_df = self.nodes['tweet']
-            duplicated = (tweet_df[tweet_df.tweet_id.duplicated()].tweet_id
-                                                                  .tolist())
-            if len(duplicated) > 0:
-                raise RuntimeError(f'The tweet IDs {duplicated} are '
-                                   f'duplicate in the dataset!')
+                # Otherwise, with underscores in the filename then we assume it
+                # contains relation data
+                else:
+                    splits = name.replace('.pickle', '').split('_')
+                    src = splits[0]
+                    tgt = splits[-1]
+                    rel = '_'.join(splits[1:-1])
+                    self.rels[(src, rel, tgt)] = df.copy()
+
+            # Ensure that claims are present in the dataset
+            if 'claim' not in self.nodes.keys():
+                raise RuntimeError('No claims are present in the file!')
+
+            # Ensure that tweets are present in the dataset, and also that the
+            # tweet IDs are unique
+            if 'tweet' not in self.nodes.keys():
+                raise RuntimeError('No tweets are present in the file!')
+            else:
+                tweet_df = self.nodes['tweet']
+                duplicated = (tweet_df[tweet_df.tweet_id.duplicated()]
+                              .tweet_id
+                              .tolist())
+                if len(duplicated) > 0:
+                    raise RuntimeError(f'The tweet IDs {duplicated} are '
+                                       f'duplicate in the dataset!')
 
         return self
 
@@ -435,7 +485,11 @@ class MuminDataset:
             return self
 
     def _rehydrate(self, node_type: str):
-        '''Rehydrate the tweets and users in the dataset'''
+        '''Rehydrate the tweets and users in the dataset.
+
+        Args:
+            node_type (str): The type of node to rehydrate.
+        '''
 
         if (node_type in self.nodes.keys() and
                 (node_type != 'reply' or self.include_replies)):
@@ -675,6 +729,14 @@ class MuminDataset:
         # Hashtags
         if self.include_hashtags:
             def extract_hashtag(dcts: List[dict]) -> List[str]:
+                '''Extracts hashtags from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[str]: A list of hashtags.
+                '''
                 return [dct.get('tag') for dct in dcts]
 
             # Add hashtags from tweets
@@ -710,6 +772,14 @@ class MuminDataset:
         # Add urls from tweets
         if 'entities.urls' in self.nodes['tweet'].columns:
             def extract_url(dcts: List[dict]) -> List[Union[str, None]]:
+                '''Extracts urls from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[Union[str, None]]: A list of urls.
+                '''
                 return [dct.get('expanded_url') or dct.get('url')
                         for dct in dcts]
             urls = (self.nodes['tweet']['entities.urls']
@@ -728,6 +798,14 @@ class MuminDataset:
         # Add urls from user urls
         if 'entities.url.urls' in self.nodes['user'].columns:
             def extract_url(dcts: List[dict]) -> List[Union[str, None]]:
+                '''Extracts urls from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[Union[str, None]]: A list of urls.
+                '''
                 return [dct.get('expanded_url') or dct.get('url')
                         for dct in dcts]
             urls = (self.nodes['user']['entities.url.urls']
@@ -746,6 +824,14 @@ class MuminDataset:
         # Add urls from user descriptions
         if 'entities.description.urls' in self.nodes['user'].columns:
             def extract_url(dcts: List[dict]) -> List[Union[str, None]]:
+                '''Extracts urls from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[Union[str, None]]: A list of urls.
+                '''
                 return [dct.get('expanded_url') or dct.get('url')
                         for dct in dcts]
             urls = (self.nodes['user']['entities.description.urls']
@@ -765,6 +851,14 @@ class MuminDataset:
         if (self.include_images and
                 'profile_image_url' in self.nodes['user'].columns):
             def extract_url(dcts: List[dict]) -> List[Union[str, None]]:
+                '''Extracts urls from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[Union[str, None]]: A list of urls.
+                '''
                 return [dct.get('expanded_url') or dct.get('url')
                         for dct in dcts]
             urls = (self.nodes['user']['profile_image_url']
@@ -833,6 +927,14 @@ class MuminDataset:
         if self.include_mentions and mentions_exist:
 
             def extract_mention(dcts: List[dict]) -> List[int]:
+                '''Extracts user ids from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[int]: A list of user ids.
+                '''
                 return [int(dct['id']) for dct in dcts]
 
             merged = (self.nodes['tweet'][['entities.mentions']]
@@ -858,6 +960,14 @@ class MuminDataset:
         if self.include_mentions and mentions_exist:
 
             def extract_mention(dcts: List[dict]) -> List[str]:
+                '''Extracts user ids from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[str]: A list of user ids.
+                '''
                 return [dct['username'] for dct in dcts]
 
             merged = (self.nodes['user'][['entities.description.mentions']]
@@ -898,6 +1008,14 @@ class MuminDataset:
         hashtags_exist = 'entities.hashtags' in self.nodes['tweet'].columns
         if self.include_hashtags and hashtags_exist:
             def extract_hashtag(dcts: List[dict]) -> List[str]:
+                '''Extracts hashtags from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[str]: A list of hashtags.
+                '''
                 return [dct.get('tag') for dct in dcts]
             merged = (self.nodes['tweet'][['entities.hashtags']]
                           .dropna()
@@ -920,6 +1038,14 @@ class MuminDataset:
         hashtags_exist = 'entities.description.hashtags' in user_cols
         if self.include_hashtags and hashtags_exist:
             def extract_hashtag(dcts: List[dict]) -> List[str]:
+                '''Extracts hashtags from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[str]: A list of hashtags.
+                '''
                 return [dct.get('tag') for dct in dcts]
             merged = (self.nodes['user'][['entities.description.hashtags']]
                           .dropna()
@@ -941,6 +1067,14 @@ class MuminDataset:
         urls_exist = 'entities.urls' in self.nodes['tweet'].columns
         if (self.include_articles or self.include_images) and urls_exist:
             def extract_url(dcts: List[dict]) -> List[Union[str, None]]:
+                '''Extracts urls from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[str]: A list of urls.
+                '''
                 return [dct.get('expanded_url') or dct.get('url')
                         for dct in dcts]
             merged = (self.nodes['tweet'][['entities.urls']]
@@ -965,6 +1099,14 @@ class MuminDataset:
         desc_urls_exist = 'entities.description.urls' in user_cols
         if self.include_images and (url_urls_exist or desc_urls_exist):
             def extract_url(dcts: List[dict]) -> List[Union[str, None]]:
+                '''Extracts urls from a list of dictionaries.
+
+                Args:
+                    dcts (List[dict]): A list of dictionaries.
+
+                Returns:
+                    List[str]: A list of urls.
+                '''
                 return [dct.get('expanded_url') or dct.get('url')
                         for dct in dcts]
 
@@ -1237,6 +1379,8 @@ class MuminDataset:
         # Throw error if `transformers` has not been installed
         try:
             import transformers  # noqa
+            from transformers import logging as tf_logging
+            tf_logging.set_verbosity_error()
         except ModuleNotFoundError:
             msg = ('You have opted to include embeddings, but you have '
                    'not installed the `transformers` library. Have you '
@@ -1246,37 +1390,55 @@ class MuminDataset:
             raise ModuleNotFoundError(msg)
 
         # Embed tweets
-        if 'tweet' in nodes_to_embed:
+        if ('tweet' in nodes_to_embed and
+                'text_emb' not in self.nodes['tweet'].columns):
             self._embed_tweets()
+            self._dump_dataset()
 
         # Embed replies
-        if 'reply' in nodes_to_embed:
+        if ('reply' in nodes_to_embed and
+                'text_emb' not in self.nodes['reply'].columns):
             self._embed_replies()
+            self._dump_dataset()
 
         # Embed users
-        if 'user' in nodes_to_embed:
+        if ('user' in nodes_to_embed and
+                'description_emb' not in self.nodes['user'].columns):
             self._embed_users()
+            self._dump_dataset()
 
         # Embed articles
-        if 'article' in nodes_to_embed:
+        if ('article' in nodes_to_embed and
+                'content_emb' not in self.nodes['article'].columns):
             self._embed_articles()
+            self._dump_dataset()
 
         # Embed images
-        if 'image' in nodes_to_embed:
+        if ('image' in nodes_to_embed and
+                'pixels_emb' not in self.nodes['image'].columns):
             self._embed_images()
+            self._dump_dataset()
 
         # Embed claims
-        if 'claim' in nodes_to_embed:
+        if ('claim' in nodes_to_embed and
+                'reviewer_emb' not in self.nodes['claim'].columns):
             self._embed_claims()
-
-        # Dump the nodes with all the embeddings
-        self._dump_dataset()
+            self._dump_dataset()
 
         return self
 
     @staticmethod
-    def _embed_text(text: str, tokenizer, model):
-        '''Extract a text embedding'''
+    def _embed_text(text: str, tokenizer, model) -> np.ndarray:
+        '''Extract a text embedding.
+
+        Args:
+            text (str): The text to embed.
+            tokenizer (transformers.PreTrainedTokenizer): The tokenizer to use.
+            model (transformers.PreTrainedModel): The model to use.
+
+        Returns:
+            np.ndarray: The embedding of the text.
+        '''
         import torch
         with torch.no_grad():
             inputs = tokenizer(text, truncation=True, return_tensors='pt')
@@ -1380,7 +1542,7 @@ class MuminDataset:
                     return self._embed_text(text, **params)
                 else:
                     return np.mean([self._embed_text(doc, **params)
-                                    for doc in text])
+                                    for doc in text], axis=0)
 
             def split_content(doc: str) -> List[str]:
                 '''Split up a string into smaller chunks'''
@@ -1431,7 +1593,7 @@ class MuminDataset:
                     # Get the embedding
                     outputs = model(**inputs, output_hidden_states=True)
                     penultimate_embedding = outputs.hidden_states[-1]
-                    cls_embedding = penultimate_embedding[:, 0, :]
+                    cls_embedding = penultimate_embedding[0, 0, :]
 
                     # Convert to NumPy and return
                     return cls_embedding.numpy()
@@ -1454,6 +1616,14 @@ class MuminDataset:
         if isinstance(self.nodes['claim'].reviewers.iloc[0], str):
 
             def string_to_list(string: str) -> list:
+                '''Convert a string to a list.
+
+                Args:
+                    string: A string to be converted to a list.
+
+                Returns:
+                    list: A list of strings.
+                '''
                 string = string.replace('\'', '\"')
                 return json.loads(string)
 
@@ -1469,7 +1639,14 @@ class MuminDataset:
                         for reviewer, array in zip(reviewers, one_hotted)}
 
         def embed_reviewers(revs: List[str]) -> np.ndarray:
-            '''One-hot encoding of multiple reviewers'''
+            '''One-hot encoding of multiple reviewers.
+
+            Args:
+                revs: A list of reviewers.
+
+            Returns:
+                np.ndarray: A one-hot encoded array.
+            '''
             arrays = [one_hot_dict[rev] for rev in revs]
             return np.stack(arrays, axis=0).sum(axis=0)
 
@@ -1486,6 +1663,8 @@ class MuminDataset:
         # Set up the node features that should be kept
         size = 'small' if self.size == 'test' else self.size
         node_feats = dict(claim=['embedding', 'label', 'reviewers', 'date',
+                                 'language', 'keywords', 'cluster_keywords',
+                                 'cluster',
                                  f'{size}_train_mask',
                                  f'{size}_val_mask',
                                  f'{size}_test_mask'],
@@ -1594,10 +1773,71 @@ class MuminDataset:
         for rel_type in rels_to_pop:
             self.rels.pop(rel_type)
 
+    def _set_datatypes(self):
+        '''Set datatypes in the dataframes, to use less memory'''
+
+        # Set up all the dtypes of the columns
+        dtypes = dict(tweet=dict(tweet_id='uint64',
+                                 text='str',
+                                 created_at={'created_at': 'datetime64[ns]'},
+                                 lang='category',
+                                 source='str',
+                                 num_retweets='uint64',
+                                 num_replies='uint64',
+                                 num_quote_tweets='uint64'),
+                      user=dict(user_id='uint64',
+                                verified='bool',
+                                protected='bool',
+                                created_at={'created_at': 'datetime64[ns]'},
+                                username='str',
+                                description='str',
+                                url='str',
+                                name='str',
+                                num_followers='uint64',
+                                num_followees='uint64',
+                                num_tweets='uint64',
+                                num_listed='uint64',
+                                location='category'),
+                      image=dict(url='str',
+                                 pixels='numpy',
+                                 width='uint64',
+                                 height='uint64'),
+                      article=dict(url='str',
+                                   title='str',
+                                   content='str'),
+                      hashtag=dict(tag='str'),
+                      reply=dict(tweet_id='uint64',
+                                 text='str',
+                                 created_at={'created_at': 'datetime64[ns]'},
+                                 lang='category',
+                                 source='str',
+                                 num_retweets='uint64',
+                                 num_replies='uint64',
+                                 num_quote_tweets='uint64'))
+
+        # Loop over all nodes
+        for ntype, dtype_dict in dtypes.items():
+            if ntype in self.nodes.keys():
+
+                # Set the dtypes for non-numpy columns
+                dtype_dict_no_numpy = {col: dtype
+                                       for col, dtype in dtype_dict.items()
+                                       if dtype != 'numpy'}
+                for col, dtype in dtype_dict_no_numpy.items():
+                    self.nodes[ntype][col] = (self.nodes[ntype][col]
+                                              .astype(dtype))
+
+                # For numpy columns, set the type manually
+                def numpy_fn(x):
+                    return np.asarray(x)
+
+                for col, dtype in dtype_dict.items():
+                    if dtype == 'numpy':
+                        self.nodes[ntype][col] = (self.nodes[ntype][col]
+                                                      .map(numpy_fn))
+
     def _remove_auxilliaries(self):
         '''Removes node types that are not in use anymore'''
-        logger.info('Removing auxilliary nodes')
-
         # Remove auxilliary node types
         nodes_to_remove = [node_type for node_type in self.nodes.keys()
                            if node_type not in self._node_dump]
@@ -1614,8 +1854,6 @@ class MuminDataset:
 
     def _remove_islands(self):
         '''Removes nodes and relations that are not connected to anything'''
-        logger.info('Removing island nodes')
-
         # Loop over all the node types
         for node_type, node_df in self.nodes.items():
 
@@ -1691,22 +1929,55 @@ class MuminDataset:
         return self
 
     def _dump_dataset(self):
-        '''Dumps the dataset to hdf files'''
+        '''Dumps the dataset to a zip file'''
         logger.info('Dumping dataset')
-        hdf_path = self.dataset_path
 
-        with warnings.catch_warnings():
-            warnings.simplefilter('ignore', PerformanceWarning)
+        # Create a temporary pickle folder
+        temp_pickle_folder = Path('temp_pickle_folder')
+        if not temp_pickle_folder.exists():
+            temp_pickle_folder.mkdir()
 
-            # Dump the nodes
-            for node_type in self._node_dump:
-                if node_type in self.nodes.keys():
-                    self.nodes[node_type].to_hdf(hdf_path, node_type)
+        # Make temporary pickle list
+        pickle_list = list()
 
-            # Dump the relations
-            for rel_type in self._rel_dump:
-                if rel_type in self.rels.keys():
-                    self.rels[rel_type].to_hdf(hdf_path, '_'.join(rel_type))
+        # Create progress bar
+        total = len(self._node_dump) + len(self._rel_dump) + 1
+        pbar = tqdm(total=total)
+
+        # Store the nodes
+        for node_type in self._node_dump:
+            pbar.set_description(f'Storing {node_type} nodes')
+            if node_type in self.nodes.keys():
+                pickle_list.append(node_type)
+                pickle_path = temp_pickle_folder / f'{node_type}.pickle'
+                self.nodes[node_type].to_pickle(pickle_path, protocol=4)
+            pbar.update()
+
+        # Store the relations
+        for rel_type in self._rel_dump:
+            pbar.set_description(f'Storing {rel_type} relations')
+            if rel_type in self.rels.keys():
+                name = '_'.join(rel_type)
+                pickle_list.append(name)
+                pickle_path = temp_pickle_folder / f'{name}.pickle'
+                self.rels[rel_type].to_pickle(pickle_path, protocol=4)
+            pbar.update()
+
+        # Zip the nodes and relations, and save the zip file
+        with zipfile.ZipFile(self.dataset_path,
+                             mode='w',
+                             compression=zipfile.ZIP_STORED) as zip_file:
+            pbar.set_description('Dumping dataset')
+            for name in pickle_list:
+                fname = f'{name}.pickle'
+                zip_file.write(temp_pickle_folder / fname, arcname=fname)
+
+        # Remove the temporary pickle folder
+        rmtree(temp_pickle_folder)
+
+        # Final progress bar update and close it
+        pbar.update()
+        pbar.close()
 
         return self
 
